"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Project = void 0;
const path_1 = __importDefault(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const utils_1 = require("../utils");
const ExportHandler_1 = require("./ExportHandler");
const structure_1 = require("./structure");
const utils_2 = require("./utils");
class Project {
    repository;
    readme;
    homepage;
    version;
    module;
    extractor;
    baseDir;
    root;
    tsconfig;
    ignoreNamespaceMembers;
    idAcc;
    constructor({ folderPath, extractor, packageJSON, tsconfig }) {
        folderPath.pop();
        this.baseDir = folderPath[folderPath.length - 1];
        this.root = packageJSON.path;
        this.tsconfig = tsconfig;
        this.repository = (0, utils_1.getRepository)(packageJSON, extractor.settings.branchName);
        this.homepage = packageJSON.contents.homepage;
        this.version = packageJSON.contents.version;
        this.readme = (0, utils_1.getReadme)(packageJSON.path);
        let name = packageJSON.contents.name;
        if (name.includes("/"))
            name = name.split("/")[1];
        this.module = (0, structure_1.createModule)(name, [], true, this.repository && `${this.repository}/${this.baseDir}`, false);
        if (extractor.settings.entryPoints.length !== 1)
            this.module.path.push(name);
        this.extractor = extractor;
        this.idAcc = 1;
    }
    visitor(sourceFile, currentModule) {
        let sym;
        let fileName;
        if ("fileName" in sourceFile) {
            if (this.extractor.program.isSourceFileFromExternalLibrary(sourceFile) || this.extractor.program.isSourceFileDefaultLibrary(sourceFile))
                return;
            fileName = sourceFile.fileName;
            sym = this.extractor.checker.getSymbolAtLocation(sourceFile);
        }
        else {
            sym = sourceFile;
            fileName = this.resolveSymbolFileName(sym);
        }
        if (!sym || !sym.exports)
            return;
        if (this.extractor.fileCache.has(fileName))
            return;
        const isCached = this.extractor.isCachedFile(fileName);
        for (const [, val] of sym.exports) {
            if (val.name === "__export") {
                for (const decl of val.declarations) {
                    if (typescript_1.default.isExportDeclaration(decl) && decl.moduleSpecifier && typescript_1.default.isStringLiteral(decl.moduleSpecifier))
                        (0, ExportHandler_1.registerDirectReExport)(this, currentModule, decl.moduleSpecifier);
                }
            }
            else if (val.declarations && val.declarations.length) {
                if (typescript_1.default.isExportSpecifier(val.declarations[0]))
                    (0, ExportHandler_1.registerOtherExportOrReExport)(this, currentModule, val);
                else if (typescript_1.default.isNamespaceExport(val.declarations[0]))
                    (0, ExportHandler_1.registerNamespaceReExport)(this, currentModule, val);
                else {
                    const sym = this.handleSymbol(val, currentModule, isCached);
                    if (sym)
                        (0, ExportHandler_1.registerDirectExport)(fileName, currentModule, sym);
                }
            }
        }
        if (this.extractor.settings.documentImports) {
            const file = sym.valueDeclaration.getSourceFile();
            for (const childFile of file.imports) {
                const source = (0, ExportHandler_1.resolveSourceFile)(this.extractor, file.fileName, childFile.text);
                if (source)
                    this.visitor(source, this.getOrCreateModule(source.fileName));
            }
        }
        return;
    }
    getOrCreateModule(source) {
        const { dir } = path_1.default.parse(source);
        if (this.extractor.moduleCache[dir])
            return this.extractor.moduleCache[dir];
        let paths = dir.split("/");
        paths = paths.slice(paths.indexOf(this.baseDir) + 1);
        if (!paths.length) {
            this.extractor.moduleCache[dir] = this.module;
            return this.module;
        }
        let lastModule = this.module;
        const newPath = [];
        const skipped = [];
        for (const pathPart of paths) {
            if (pathPart === "")
                break;
            const newMod = lastModule.modules.get(pathPart);
            if (this.extractor.settings.passthroughModules?.includes(pathPart)) {
                skipped.push(pathPart);
                continue;
            }
            newPath.push(pathPart);
            if (!newMod) {
                let repoPath;
                if (lastModule.repository) {
                    repoPath = lastModule.repository;
                    if (skipped.length) {
                        if (!repoPath.split("/").some(p => skipped.includes(p)))
                            repoPath += `/${skipped.join("/")}`;
                        skipped.length = 0;
                    }
                    repoPath += `/${pathPart}`;
                }
                const mod = (0, structure_1.createModule)(pathPart, [...this.module.path, ...newPath], false, repoPath, false);
                lastModule.modules.set(pathPart, mod);
                lastModule = mod;
            }
            else
                lastModule = newMod;
        }
        if (skipped.length)
            lastModule.repository += `/${skipped.join("/")}`;
        this.extractor.moduleCache[dir] = lastModule;
        return lastModule;
    }
    forEachModule(cb, module = this.module) {
        return (0, utils_2.forEachModule)(module, cb);
    }
    handleSymbol(val, currentModule, isCached) {
        if (!val.declarations || !val.declarations.length)
            return;
        if (!currentModule) {
            const origin = val.declarations[0].getSourceFile();
            if (this.extractor.program.isSourceFileFromExternalLibrary(origin) || this.extractor.program.isSourceFileDefaultLibrary(origin))
                return;
            const fileName = origin.fileName;
            currentModule = this.getOrCreateModule(fileName);
            this.visitor(origin, currentModule);
            isCached = this.extractor.fileCache.has(fileName) ? this.extractor.fileCache.get(fileName) : this.extractor.settings.fileCache?.has(fileName);
        }
        if (this.extractor.refs.has(val))
            return this.extractor.refs.get(val);
        if (!this.ignoreNamespaceMembers && typescript_1.default.isModuleBlock(val.declarations[0].parent)) {
            const namespaceSym = this.extractor.checker.getSymbolAtLocation(val.declarations[0].parent.parent.name);
            if (namespaceSym) {
                if (!this.extractor.refs.has(namespaceSym))
                    this.handleNamespaceDecl(namespaceSym, currentModule, isCached);
                return this.extractor.refs.get(val);
            }
        }
        let type;
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.Module))
            type = this.handleNamespaceDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.Interface))
            type = this.handleInterfaceDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.Enum))
            type = this.handleEnumDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.TypeAlias))
            type = this.handleTypeAliasDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.Variable) && !(0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.FunctionScopedVariable))
            type = this.handleVariableDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.Function))
            type = this.handleFunctionDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.Class) && val.name !== "__class")
            type = this.handleClassDecl(val, currentModule, isCached);
        if ((0, utils_1.hasBit)(val.flags, typescript_1.default.SymbolFlags.EnumMember)) {
            const parent = val.parent;
            if (!this.extractor.refs.has(parent))
                this.handleEnumDecl(parent, currentModule, isCached);
            type = this.extractor.refs.get(val);
        }
        if (type)
            return type;
        const aliased = this.resolveAliasedSymbol(val);
        if (this.extractor.refs.has(aliased))
            return this.extractor.refs.get(aliased);
        if (aliased.declarations && aliased.declarations.length)
            return this.extractor.refs.findExternal(aliased);
        return;
    }
    handleClassDecl(symbol, currentModule, isCached) {
        const decl = symbol.declarations.find(decl => typescript_1.default.isClassDeclaration(decl));
        const name = symbol.name;
        const ref = {
            name,
            path: currentModule.path,
            kind: 0,
        };
        this.extractor.refs.set(symbol, ref);
        if (this.isInternalNode(decl)) {
            ref.kind = 12;
            return ref;
        }
        const properties = [];
        const methods = new Map();
        let constructor;
        for (const member of decl.members) {
            let isStatic, isPrivate, isProtected, isReadonly, isAbstract;
            if (member.modifiers) {
                for (const modifier of member.modifiers) {
                    if (modifier.kind === typescript_1.default.SyntaxKind.StaticKeyword)
                        isStatic = true;
                    else if (modifier.kind === typescript_1.default.SyntaxKind.ProtectedKeyword)
                        isProtected = true;
                    else if (modifier.kind === typescript_1.default.SyntaxKind.PrivateKeyword)
                        isPrivate = true;
                    else if (modifier.kind === typescript_1.default.SyntaxKind.ReadonlyKeyword)
                        isReadonly = true;
                    else if (modifier.kind === typescript_1.default.SyntaxKind.AbstractKeyword)
                        isAbstract = true;
                }
            }
            if (typescript_1.default.isIndexSignatureDeclaration(member)) {
                properties.push({
                    index: {
                        key: member.parameters[0]?.type && this.resolveType(member.parameters[0].type),
                        type: this.resolveType(member.type)
                    },
                    isStatic, isReadonly,
                    loc: this.getLOC(currentModule, member),
                    jsDoc: this.getJSDocData(member)
                });
            }
            if (typescript_1.default.isPropertyDeclaration(member)) {
                if (this.isInternalNode(member))
                    continue;
                const computedName = typescript_1.default.isComputedPropertyName(member.name) && this.resolveExpressionToType(member.name.expression);
                properties.push({
                    prop: {
                        name: computedName || member.name.text,
                        rawName: member.name.text,
                        type: member.type && this.resolveType(member.type),
                        isOptional: Boolean(member.questionToken),
                        initializer: member.initializer && this.resolveExpressionToType(member.initializer)
                    },
                    isPrivate, isProtected, isStatic, isReadonly, isAbstract,
                    jsDoc: this.getJSDocData(member),
                    loc: this.getLOC(currentModule, member)
                });
            }
            else if (typescript_1.default.isConstructorDeclaration(member)) {
                if (this.isInternalNode(member))
                    continue;
                if (!constructor) {
                    constructor = {
                        loc: this.getLOC(currentModule, member),
                        signatures: [{
                                parameters: member.parameters.map(p => this.resolveParameter(p)),
                                jsDoc: this.getJSDocData(member)
                            }],
                    };
                }
                else {
                    constructor.signatures.push({ parameters: member.parameters.map(p => this.resolveParameter(p)), jsDoc: this.getJSDocData(member) });
                    if (member.body)
                        constructor.loc = this.getLOC(currentModule, member);
                }
            }
            else if (typescript_1.default.isMethodDeclaration(member)) {
                if (this.isInternalNode(member))
                    continue;
                const methodName = member.name.getText();
                const method = methods.get(methodName);
                const computedName = typescript_1.default.isComputedPropertyName(member.name) ? this.resolveExpressionToType(member.name.expression) : undefined;
                if (method) {
                    method.signatures.push({
                        returnType: this.resolveReturnType(member),
                        typeParameters: member.typeParameters && member.typeParameters.map(p => this.resolveTypeParameters(p)),
                        parameters: member.parameters.map(p => this.resolveParameter(p)),
                        jsDoc: this.getJSDocData(member)
                    });
                    if (member.body)
                        method.loc = this.getLOC(currentModule, member);
                }
                else {
                    methods.set(methodName, {
                        name: computedName || methodName,
                        rawName: methodName,
                        loc: this.getLOC(currentModule, member),
                        isPrivate, isProtected, isStatic, isAbstract,
                        jsDoc: this.getJSDocData(member),
                        isGenerator: Boolean(member.asteriskToken),
                        signatures: [{
                                returnType: this.resolveReturnType(member),
                                typeParameters: member.typeParameters && member.typeParameters.map(p => this.resolveTypeParameters(p)),
                                parameters: member.parameters.map(p => this.resolveParameter(p)),
                                jsDoc: this.getJSDocData(member)
                            }]
                    });
                }
            }
            else if (typescript_1.default.isGetAccessor(member)) {
                if (this.isInternalNode(member))
                    continue;
                const methodName = member.name.getText();
                const computedName = typescript_1.default.isComputedPropertyName(member.name) ? this.resolveExpressionToType(member.name.expression) : undefined;
                methods.set(methodName, {
                    name: computedName || methodName,
                    rawName: methodName,
                    signatures: [{
                            returnType: this.resolveReturnType(member),
                            jsDoc: this.getJSDocData(member)
                        }],
                    isPrivate, isProtected, isStatic, isAbstract,
                    loc: this.getLOC(currentModule, member),
                    jsDoc: this.getJSDocData(member),
                    isGetter: true
                });
            }
            else if (typescript_1.default.isSetAccessor(member)) {
                if (this.isInternalNode(member))
                    continue;
                const methodName = member.name.getText();
                const computedName = typescript_1.default.isComputedPropertyName(member.name) ? this.resolveExpressionToType(member.name.expression) : undefined;
                methods.set(methodName, {
                    name: computedName || methodName,
                    rawName: methodName,
                    signatures: [{
                            returnType: this.resolveReturnType(member),
                            parameters: member.parameters.map(p => this.resolveParameter(p)),
                            jsDoc: this.getJSDocData(member)
                        }],
                    isPrivate, isProtected, isStatic, isAbstract,
                    loc: this.getLOC(currentModule, member),
                    jsDoc: this.getJSDocData(member),
                    isSetter: true
                });
            }
        }
        const classObj = {
            name,
            typeParameters: decl.typeParameters?.map(p => this.resolveTypeParameters(p)),
            properties,
            methods: [...methods.values()],
            loc: this.getLOC(currentModule, decl, true),
            jsDoc: this.getJSDocData(decl),
            isAbstract: decl.modifiers && decl.modifiers.some(m => m.kind === typescript_1.default.SyntaxKind.AbstractKeyword),
            _constructor: constructor,
            isCached,
            kind: 0
        };
        if (decl.heritageClauses) {
            const extendsClause = decl.heritageClauses.find(clause => clause.token === typescript_1.default.SyntaxKind.ExtendsKeyword);
            classObj.extends = extendsClause && this.resolveHeritage(extendsClause.types[0]);
            const implementsClauses = decl.heritageClauses?.find(clause => clause.token === typescript_1.default.SyntaxKind.ImplementsKeyword);
            classObj.implements = implementsClauses && implementsClauses.types.map(clause => this.resolveHeritage(clause));
        }
        if (currentModule.classes.some(cl => cl.name === name))
            classObj.id = ref.id = this.idAcc++;
        currentModule.classes.push(classObj);
        return ref;
    }
    handleInterfaceDecl(sym, currentModule, isCached) {
        const firstDecl = sym.declarations.find(decl => typescript_1.default.isInterfaceDeclaration(decl));
        const ref = {
            name: sym.name,
            path: currentModule.path,
            kind: 1
        };
        this.extractor.refs.set(sym, ref);
        if (this.isInternalNode(firstDecl)) {
            ref.kind = 12;
            return ref;
        }
        const properties = [];
        const loc = [];
        const jsDoc = [];
        for (const decl of sym.declarations) {
            if (!typescript_1.default.isInterfaceDeclaration(decl))
                continue;
            for (const member of (decl.members || [])) {
                if (this.isInternalNode(member))
                    continue;
                properties.push(this.resolveObjectProperty(member));
            }
            loc.push(this.getLOC(currentModule, decl, true));
            const jsdoc = this.getJSDocData(decl);
            if (jsdoc)
                jsDoc.push(...jsdoc);
        }
        let extendsInt, implementsInt;
        if (firstDecl.heritageClauses) {
            const extendsClause = firstDecl.heritageClauses.find(c => c.token === typescript_1.default.SyntaxKind.ExtendsKeyword);
            extendsInt = extendsClause && extendsClause.types.map(t => this.resolveHeritage(t));
            const implementsClause = firstDecl.heritageClauses.find(c => c.token === typescript_1.default.SyntaxKind.ImplementsKeyword);
            implementsInt = implementsClause && implementsClause.types.map(impl => this.resolveType(impl));
        }
        let id;
        if (currentModule.interfaces.some(int => int.name === sym.name))
            ref.id = id = this.idAcc++;
        currentModule.interfaces.push({
            name: sym.name,
            extends: extendsInt,
            implements: implementsInt,
            loc,
            properties,
            jsDoc,
            id,
            typeParameters: firstDecl.typeParameters && firstDecl.typeParameters.map(p => this.resolveTypeParameters(p)),
            isCached: (isCached && sym.declarations.length === 1) ? true : undefined,
            kind: 1
        });
        return ref;
    }
    handleEnumDecl(sym, currentModule, isCached) {
        const firstDecl = sym.declarations.find(decl => typescript_1.default.isEnumDeclaration(decl));
        const ref = {
            name: sym.name,
            path: currentModule.path,
            kind: 2
        };
        this.extractor.refs.set(sym, ref);
        if (this.isInternalNode(firstDecl)) {
            ref.kind = 12;
            return ref;
        }
        const members = [];
        const loc = [];
        const jsDoc = [];
        for (const decl of sym.declarations) {
            if (!typescript_1.default.isEnumDeclaration(decl))
                continue;
            for (const el of (decl.members || [])) {
                const isInternal = this.isInternalNode(el);
                const name = el.name.getText();
                if (!isInternal)
                    members.push({
                        name,
                        initializer: el.initializer && this.resolveExpressionToType(el.initializer, false),
                        loc: this.getLOC(currentModule, el),
                        jsDoc: this.getJSDocData(el)
                    });
                const elSymbol = this.extractor.checker.getSymbolAtLocation(el.name);
                if (elSymbol) {
                    this.extractor.refs.set(elSymbol, {
                        name: sym.name,
                        displayName: name,
                        path: currentModule.path,
                        kind: isInternal ? 12 : 9
                    });
                }
            }
            loc.push(this.getLOC(currentModule, decl, true));
            const jsDocData = this.getJSDocData(decl);
            if (jsDocData)
                jsDoc.push(...jsDocData);
        }
        let id;
        if (currentModule.enums.some(int => int.name === sym.name))
            ref.id = id = this.idAcc++;
        currentModule.enums.push({
            name: sym.name,
            isConst: Boolean(firstDecl.modifiers && firstDecl.modifiers.some(mod => mod.kind === typescript_1.default.SyntaxKind.ConstKeyword)),
            loc,
            jsDoc,
            members,
            id,
            isCached: (isCached && sym.declarations.length === 1) ? true : undefined,
            kind: 2
        });
        return ref;
    }
    handleTypeAliasDecl(sym, currentModule, isCached) {
        const decl = sym.declarations.find(decl => typescript_1.default.isTypeAliasDeclaration(decl));
        const ref = {
            name: sym.name,
            path: currentModule.path,
            kind: 5
        };
        this.extractor.refs.set(sym, ref);
        if (this.isInternalNode(decl)) {
            ref.kind = 12;
            return ref;
        }
        let id;
        if (currentModule.types.some(int => int.name === sym.name))
            ref.id = id = this.idAcc++;
        currentModule.types.push({
            name: sym.name,
            value: this.resolveType(decl.type),
            typeParameters: decl.typeParameters?.map(param => this.resolveTypeParameters(param)),
            loc: this.getLOC(currentModule, decl, true),
            jsDoc: this.getJSDocData(decl),
            id,
            isCached,
            kind: 5
        });
        return ref;
    }
    handleNamespaceDecl(symbol, currentModule, isCached) {
        const firstDecl = symbol.declarations.find(t => typescript_1.default.isModuleDeclaration(t));
        if (!firstDecl)
            return;
        const newMod = (0, structure_1.createModule)(firstDecl.name.text, [...currentModule.path, firstDecl.name.text], false, this.getLOC(currentModule, firstDecl).sourceFile, true);
        newMod.exports.index = { exports: [], reExports: [] };
        const namespaceLoc = this.getLOC(currentModule, firstDecl);
        newMod.repository = namespaceLoc.sourceFile;
        currentModule.modules.set(newMod.name, newMod);
        newMod.jsDoc = this.getJSDocData(firstDecl);
        const ref = {
            name: symbol.name,
            path: currentModule.path,
            kind: 10
        };
        this.extractor.refs.set(symbol, ref);
        this.ignoreNamespaceMembers = true;
        const areChildrenCached = isCached && symbol.declarations.length === 1 ? true : undefined;
        for (const [, element] of symbol.exports) {
            if (!(0, utils_1.hasBit)(element.flags, typescript_1.default.SymbolFlags.ModuleMember))
                continue;
            const sym = this.handleSymbol(element, newMod, areChildrenCached);
            if (sym)
                newMod.exports.index.exports.push(sym);
        }
        this.ignoreNamespaceMembers = false;
        return ref;
    }
    handleVariableDecl(sym, currentModule, isCached) {
        const decl = sym.declarations.find(decl => typescript_1.default.isVariableDeclaration(decl) || typescript_1.default.isBindingElement(decl));
        const ref = {
            name: sym.name,
            kind: 4,
            path: currentModule.path,
        };
        this.extractor.refs.set(sym, ref);
        if (this.isInternalNode(decl)) {
            ref.kind = 12;
            return ref;
        }
        let name = "";
        let type;
        let comment;
        let text = "";
        let id;
        if (typescript_1.default.isBindingElement(decl)) {
            type = this.resolveTypeType(this.extractor.checker.getTypeAtLocation(decl));
            name = decl.name.getText();
            comment = this.getJSDocDataRaw(decl);
        }
        else if (typescript_1.default.isVariableDeclaration(decl)) {
            const maxLen = this.extractor.settings.maxConstantTextLength || 256;
            type = decl.type && this.resolveType(decl.type);
            if (decl.initializer) {
                const initializerText = decl.initializer.getText();
                text = initializerText && (initializerText.length > maxLen) ? `${initializerText.slice(0, maxLen)}...` : initializerText;
                if (!type)
                    type = this.resolveExpressionToType(decl.initializer);
            }
            if (currentModule.constants.some(int => int.name === sym.name))
                ref.id = id = this.idAcc++;
            name = decl.name.getText();
            comment = this.getJSDocData(decl.parent.parent);
        }
        const realType = type;
        if (realType.kind === 1) {
            ref.kind = 3;
            currentModule.functions.push({
                name,
                loc: this.getLOC(currentModule, decl, true),
                signatures: [{
                        parameters: realType.parameters,
                        returnType: realType.returnType,
                        typeParameters: realType.typeParameters
                    }],
                jsDoc: comment,
                kind: 3
            });
        }
        else
            currentModule.constants.push({
                name,
                loc: this.getLOC(currentModule, decl, true),
                jsDoc: comment,
                content: text,
                type,
                id,
                isCached,
                kind: 4
            });
        return ref;
    }
    handleFunctionDecl(sym, currentModule, isCached) {
        const lastDecl = sym.declarations[sym.declarations.length - 1];
        const ref = {
            name: sym.name,
            path: currentModule.path,
            kind: 3
        };
        this.extractor.refs.set(sym, ref);
        if (this.isInternalNode(lastDecl)) {
            ref.kind = 12;
            return ref;
        }
        const signatures = [];
        for (const decl of sym.declarations) {
            if (!typescript_1.default.isFunctionDeclaration(decl))
                continue;
            signatures.push({
                returnType: this.resolveReturnType(decl),
                typeParameters: decl.typeParameters?.map(param => this.resolveTypeParameters(param)),
                parameters: decl.parameters.map(param => this.resolveParameter(param)),
                jsDoc: this.getJSDocData(decl)
            });
        }
        let id;
        if (currentModule.functions.some(int => int.name === sym.name))
            ref.id = id = this.idAcc++;
        currentModule.functions.push({
            name: sym.name,
            signatures,
            loc: this.getLOC(currentModule, lastDecl, true),
            isGenerator: Boolean(lastDecl.asteriskToken),
            id,
            isCached,
            kind: 3
        });
        return ref;
    }
    resolveSymbolOrStr(node, typeArguments) {
        const expSym = this.extractor.checker.getSymbolAtLocation(node);
        if (expSym && expSym.name === "unknown" && typescript_1.default.isQualifiedName(node)) {
            const leftSym = this.extractor.checker.getSymbolAtLocation(node.left);
            if (leftSym) {
                const external = this.extractor.refs.findExternal(leftSym, undefined, node.right.text);
                if (external)
                    return { kind: 0, type: external, typeArguments };
            }
            return { kind: 9, name: node.getText() };
        }
        if (!expSym) {
            if (!node.getSourceFile())
                return { kind: 8 };
            const external = this.extractor.refs.findUnnamedExternal(node.getText());
            if (external)
                return { kind: 0, type: external, typeArguments };
            return { kind: 9, name: node.getText() };
        }
        if (expSym.name === "unknown")
            return { kind: 9, name: node.getText() };
        return this.resolveSymbol(expSym, typeArguments);
    }
    resolveSymbol(sym, typeArguments) {
        sym = this.resolveAliasedSymbol(sym);
        if ((0, utils_1.hasBit)(sym.flags, typescript_1.default.SymbolFlags.TypeParameter))
            return {
                type: { name: sym.name, kind: 6 },
                typeArguments,
                kind: 0
            };
        if (this.extractor.refs.has(sym))
            return {
                kind: 0,
                typeArguments,
                type: this.extractor.refs.get(sym)
            };
        const newlyCreated = this.handleSymbol(sym, undefined);
        if (newlyCreated)
            return { kind: 0, typeArguments, type: newlyCreated };
        const possiblyExternal = this.extractor.refs.findExternal(sym);
        if (possiblyExternal)
            return { kind: 0, typeArguments, type: possiblyExternal };
        return { kind: 0, typeArguments, type: { kind: 7, name: sym.name } };
    }
    resolveType(type) {
        if (typescript_1.default.isTypeReferenceNode(type)) {
            if ("symbol" in type.typeName)
                return this.resolveSymbol(type.typeName.symbol, type.typeArguments?.map(arg => this.resolveType(arg)));
            return this.resolveSymbolOrStr(type.typeName, type.typeArguments?.map(arg => this.resolveType(arg)));
        }
        else if (typescript_1.default.isFunctionTypeNode(type)) {
            return {
                typeParameters: type.typeParameters && type.typeParameters.map(p => this.resolveTypeParameters(p)),
                returnType: this.resolveType(type.type),
                parameters: type.parameters.map(p => this.resolveParameter(p)),
                kind: 1
            };
        }
        else if (typescript_1.default.isTypeLiteralNode(type)) {
            return {
                properties: type.members.map(p => this.resolveObjectProperty(p)),
                kind: 2
            };
        }
        else if (typescript_1.default.isUnionTypeNode(type)) {
            return {
                types: type.types.map(t => this.resolveType(t)),
                kind: 4
            };
        }
        else if (typescript_1.default.isIntersectionTypeNode(type)) {
            return {
                types: type.types.map(t => this.resolveType(t)),
                kind: 11
            };
        }
        else if (typescript_1.default.isTupleTypeNode(type)) {
            return {
                types: type.elements.map(el => {
                    const jsDoc = this.getJSDocDataRaw(el);
                    if (typescript_1.default.isNamedTupleMember(el))
                        return { type: this.resolveType(el.type), name: el.name.text, spread: Boolean(el.dotDotDotToken), optional: Boolean(el.questionToken), jsDoc };
                    else if (typescript_1.default.isRestTypeNode(el))
                        return { type: this.resolveType(el.type), spread: true, jsDoc };
                    else if (typescript_1.default.isOptionalTypeNode(el))
                        return { type: this.resolveType(el.type), optional: true, jsDoc };
                    return { type: this.resolveType(el), jsDoc };
                }),
                kind: 3
            };
        }
        else if (typescript_1.default.isTypePredicateNode(type)) {
            if (typescript_1.default.isThisTypeNode(type.parameterName))
                return { kind: 30, parameter: { kind: 31 }, type: type.type && this.resolveType(type.type) };
            else
                return {
                    kind: 30,
                    parameter: type.parameterName.text,
                    type: type.type && this.resolveType(type.type)
                };
        }
        else if (typescript_1.default.isTypeOperatorNode(type)) {
            let kind;
            switch (type.operator) {
                case typescript_1.default.SyntaxKind.UniqueKeyword:
                    kind = 5;
                    break;
                case typescript_1.default.SyntaxKind.KeyOfKeyword:
                    kind = 7;
                    break;
                case typescript_1.default.SyntaxKind.ReadonlyKeyword:
                    kind = 6;
            }
            return {
                kind,
                type: this.resolveType(type.type)
            };
        }
        else if (typescript_1.default.isArrayTypeNode(type)) {
            return {
                type: this.resolveType(type.elementType),
                kind: 10
            };
        }
        else if (typescript_1.default.isInferTypeNode(type)) {
            return {
                kind: 34,
                typeParameter: { kind: 0, type: { kind: 6, name: type.typeParameter.name.text } }
            };
        }
        else if (typescript_1.default.isParenthesizedTypeNode(type))
            return this.resolveType(type.type);
        else if (typescript_1.default.isThisTypeNode(type)) {
            const sym = this.extractor.checker.getSymbolAtLocation(type);
            if (!sym)
                return { name: "this", kind: 9 };
            return this.resolveSymbol(sym);
        }
        else if (typescript_1.default.isMappedTypeNode(type)) {
            return {
                typeParameter: type.typeParameter.name.text,
                optional: type.questionToken ? type.questionToken.kind === typescript_1.default.SyntaxKind.MinusToken ? 1 : 0 : undefined,
                readonly: type.readonlyToken ? type.readonlyToken.kind === typescript_1.default.SyntaxKind.MinusToken ? 1 : 0 : undefined,
                type: type.type && this.resolveType(type.type),
                constraint: type.typeParameter.constraint && this.resolveType(type.typeParameter.constraint),
                as: type.nameType && this.resolveType(type.nameType),
                kind: 23
            };
        }
        else if (typescript_1.default.isConditionalTypeNode(type)) {
            return {
                checkType: this.resolveType(type.checkType),
                extendsType: this.resolveType(type.extendsType),
                trueType: this.resolveType(type.trueType),
                falseType: this.resolveType(type.falseType),
                kind: 24
            };
        }
        else if (typescript_1.default.isTemplateLiteralTypeNode(type)) {
            return {
                head: type.head.text,
                spans: type.templateSpans.map(sp => ({ type: this.resolveType(sp.type), text: sp.literal.text })),
                kind: 25
            };
        }
        else if (typescript_1.default.isIndexedAccessTypeNode(type)) {
            return {
                object: this.resolveType(type.objectType),
                index: this.resolveType(type.indexType),
                kind: 26
            };
        }
        else if (typescript_1.default.isTypeQueryNode(type)) {
            return {
                type: this.resolveSymbolOrStr(type.exprName),
                kind: 27
            };
        }
        else if (typescript_1.default.isConstructorTypeNode(type)) {
            return {
                kind: 36,
                returnType: type.type && this.resolveType(type.type),
                parameters: type.parameters?.map(param => this.resolveParameter(param)),
                typeParameters: type.typeParameters?.map(param => this.resolveTypeParameters(param))
            };
        }
        else
            switch (type.kind) {
                case typescript_1.default.SyntaxKind.LiteralType: return this.resolveType(type.literal);
                case typescript_1.default.SyntaxKind.NumberKeyword: return { kind: 12 };
                case typescript_1.default.SyntaxKind.StringKeyword: return { kind: 13 };
                case typescript_1.default.SyntaxKind.BooleanKeyword: return { kind: 14 };
                case typescript_1.default.SyntaxKind.TrueKeyword: return { kind: 16 };
                case typescript_1.default.SyntaxKind.FalseKeyword: return { kind: 17 };
                case typescript_1.default.SyntaxKind.UndefinedKeyword: return { kind: 18 };
                case typescript_1.default.SyntaxKind.NullKeyword: return { kind: 19 };
                case typescript_1.default.SyntaxKind.VoidKeyword: return { kind: 15 };
                case typescript_1.default.SyntaxKind.AnyKeyword: return { kind: 20 };
                case typescript_1.default.SyntaxKind.UnknownKeyword: return { kind: 8 };
                case typescript_1.default.SyntaxKind.BigIntLiteral:
                case typescript_1.default.SyntaxKind.PrefixUnaryExpression:
                case typescript_1.default.SyntaxKind.PostfixUnaryExpression:
                case typescript_1.default.SyntaxKind.NumericLiteral: return { name: type.text, kind: 21 };
                case typescript_1.default.SyntaxKind.StringLiteral: return { name: type.text, kind: 22 };
                case typescript_1.default.SyntaxKind.RegularExpressionLiteral: return { name: type.text, kind: 35 };
                case typescript_1.default.SyntaxKind.SymbolKeyword: return { kind: 28 };
                case typescript_1.default.SyntaxKind.BigIntKeyword: return { kind: 29 };
                case typescript_1.default.SyntaxKind.NeverKeyword: return { kind: 32 };
                case typescript_1.default.SyntaxKind.ObjectKeyword: return { kind: 33 };
                default: return { name: type.pos === -1 ? "unknown" : type.getText(), kind: 9 };
            }
    }
    resolveReturnType(fn) {
        if (fn.type)
            return this.resolveType(fn.type);
        const sig = this.extractor.checker.getSignatureFromDeclaration(fn);
        if (!sig)
            return;
        return this.resolveTypeType(sig.getReturnType());
    }
    resolveTypeParameters(generic) {
        return {
            name: generic.name.text,
            default: generic.default ? this.resolveType(generic.default) : undefined,
            constraint: generic.constraint ? this.resolveType(generic.constraint) : undefined
        };
    }
    resolveObjectProperty(prop) {
        if (typescript_1.default.isPropertySignature(prop)) {
            const computedName = typescript_1.default.isComputedPropertyName(prop.name) && this.resolveExpressionToType(prop.name.expression);
            return {
                prop: {
                    name: computedName || prop.name.text,
                    rawName: prop.name.text,
                    type: prop.type && this.resolveType(prop.type),
                    isOptional: Boolean(prop.questionToken),
                    isReadonly: prop.modifiers?.some(mod => mod.kind === typescript_1.default.SyntaxKind.ReadonlyKeyword)
                },
                jsDoc: this.getJSDocDataRaw(prop)
            };
        }
        else if (typescript_1.default.isMethodSignature(prop)) {
            const computedName = typescript_1.default.isComputedPropertyName(prop.name) && this.resolveExpressionToType(prop.name.expression);
            return {
                prop: {
                    name: computedName || prop.name.text,
                    rawName: prop.name.text,
                    type: {
                        kind: 1,
                        parameters: prop.parameters?.map(param => this.resolveParameter(param)),
                        typeParameters: prop.typeParameters?.map(param => this.resolveTypeParameters(param)),
                        returnType: prop.type && this.resolveType(prop.type)
                    },
                    isOptional: Boolean(prop.questionToken),
                },
                jsDoc: this.getJSDocDataRaw(prop)
            };
        }
        else if (typescript_1.default.isCallSignatureDeclaration(prop))
            return {
                call: {
                    parameters: prop.parameters?.map(param => this.resolveParameter(param)),
                    typeParameters: prop.typeParameters?.map(param => this.resolveTypeParameters(param)),
                    returnType: prop.type && this.resolveType(prop.type),
                },
                jsDoc: this.getJSDocDataRaw(prop)
            };
        else if (typescript_1.default.isConstructSignatureDeclaration(prop))
            return {
                construct: {
                    parameters: prop.parameters.map(p => this.resolveParameter(p)),
                    typeParameters: prop.typeParameters?.map(param => this.resolveTypeParameters(param)),
                    returnType: prop.type && this.resolveType(prop.type),
                },
                jsDoc: this.getJSDocDataRaw(prop)
            };
        else {
            const param = prop.parameters[0];
            return {
                index: {
                    key: param.type && this.resolveType(param.type),
                    type: this.resolveType(prop.type),
                    isReadonly: prop.modifiers?.some(mod => mod.kind === typescript_1.default.SyntaxKind.ReadonlyKeyword)
                },
                jsDoc: this.getJSDocDataRaw(prop)
            };
        }
    }
    resolveHeritage(param) {
        return this.resolveSymbolOrStr(param.expression, param.typeArguments?.map(arg => this.resolveType(arg)));
    }
    resolveParameter(param) {
        const name = typescript_1.default.isIdentifier(param.name) ? param.name.text : "__namedParameters";
        return {
            name,
            isOptional: Boolean(param.questionToken),
            rest: Boolean(param.dotDotDotToken),
            type: param.type && this.resolveType(param.type),
            defaultValue: param.initializer && this.resolveExpressionToType(param.initializer),
            description: this.getJSDocCommentOfParam(param),
            jsDoc: this.getJSDocDataRaw(param)?.[0]
        };
    }
    resolveExpressionToType(exp, tryType = true) {
        if (typescript_1.default.isNewExpression(exp))
            return this.resolveSymbolOrStr(exp.expression, exp.typeArguments?.map(arg => this.resolveType(arg)));
        switch (exp.kind) {
            case typescript_1.default.SyntaxKind.BigIntLiteral:
            case typescript_1.default.SyntaxKind.PrefixUnaryExpression:
            case typescript_1.default.SyntaxKind.PostfixUnaryExpression:
            case typescript_1.default.SyntaxKind.NumericLiteral: return { name: exp.getText(), kind: 21 };
            case typescript_1.default.SyntaxKind.FalseKeyword: return { kind: 17 };
            case typescript_1.default.SyntaxKind.TrueKeyword: return { kind: 16 };
            case typescript_1.default.SyntaxKind.StringLiteral: return { name: exp.text, kind: 22 };
            case typescript_1.default.SyntaxKind.NullKeyword: return { kind: 19 };
            case typescript_1.default.SyntaxKind.RegularExpressionLiteral: return { name: exp.getText(), kind: 35 };
            case typescript_1.default.SyntaxKind.UndefinedKeyword: return { kind: 18 };
            case typescript_1.default.SyntaxKind.ArrowFunction: return {
                kind: 1,
                parameters: exp.parameters.map(p => ({
                    name: typescript_1.default.isIdentifier(p.name) ? p.name.text : "__namedParameters",
                    rest: Boolean(p.dotDotDotToken),
                    isOptional: Boolean(p.questionToken),
                    defaultValue: p.initializer && this.resolveExpressionToType(p.initializer),
                    type: p.type && this.resolveType(p.type)
                })),
                returnType: this.resolveReturnType(exp),
                typeParameters: exp.typeParameters && exp.typeParameters.map(t => this.resolveTypeParameters(t))
            };
            default: {
                if (tryType) {
                    const type = this.extractor.checker.getTypeAtLocation(exp);
                    if (!type)
                        return { kind: 9, name: exp.getText() };
                    if (type.symbol) {
                        const aliased = this.resolveAliasedSymbol(type.symbol);
                        if (this.extractor.refs.has(aliased))
                            return {
                                kind: 0,
                                type: this.extractor.refs.get(aliased),
                                typeArguments: this.extractor.checker.getTypeArguments(type)?.map(arg => this.resolveTypeType(arg))
                            };
                    }
                    const res = this.resolveTypeType(type);
                    if (res.kind === 8 || res.type?.kind === 7)
                        return { kind: 9, name: exp.getText() };
                    return res;
                }
                return { kind: 9, name: exp.getText() };
            }
        }
    }
    resolveTypeType(type) {
        if (type.isStringLiteral())
            return { kind: 22, name: type.value };
        else if (type.isNumberLiteral())
            return { kind: 21, name: type.value.toString() };
        else if (type.isUnion()) {
            const types = type.types.map(v => this.resolveTypeType(v));
            if (types[0].kind === 17 && types[1].kind === 16)
                return { kind: 14 };
            return { kind: 4, types };
        }
        else if (type.isIntersection())
            return { kind: 11, types: type.types.map(v => this.resolveTypeType(v)) };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Unknown))
            return { kind: 8 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.String))
            return { kind: 13 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Boolean))
            return { kind: 14 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Number))
            return { kind: 12 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Undefined))
            return { kind: 18 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Null))
            return { kind: 19 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Void))
            return { kind: 15 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Never))
            return { kind: 32 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.Any))
            return { kind: 20 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.ESSymbol))
            return { kind: 28 };
        else if ((0, utils_1.hasBit)(type.flags, typescript_1.default.TypeFlags.BigIntLike))
            return { kind: 29 };
        const typeNode = this.extractor.checker.typeToTypeNode(type, undefined, undefined);
        if (typeNode) {
            if (typeNode.kind === typescript_1.default.SyntaxKind.ArrayType)
                return { kind: 10, type: this.resolveTypeType(this.extractor.checker.getTypeArguments(type)[0]) };
            else if (typeNode.kind === typescript_1.default.SyntaxKind.TupleType)
                return this.resolveType(typeNode);
        }
        if (type.symbol && type.symbol.name === "__object") {
            const properties = [];
            for (const property of type.getProperties()) {
                properties.push({
                    prop: {
                        name: property.name,
                        rawName: property.name,
                        type: this.resolveTypeType(this.extractor.checker.getTypeOfSymbolAtLocation(property, property.valueDeclaration))
                    },
                    jsDoc: property.declarations?.length ? this.getJSDocDataRaw(property.declarations[0]) : undefined
                });
            }
            if (type.getNumberIndexType()) {
                properties.push({
                    index: {
                        key: { kind: 12 },
                        type: this.resolveTypeType(type.getNumberIndexType())
                    }
                });
            }
            if (type.getStringIndexType()) {
                properties.push({
                    index: {
                        key: { kind: 13 },
                        type: this.resolveTypeType(type.getStringIndexType())
                    }
                });
            }
            return {
                kind: 2,
                properties
            };
        }
        else if (type.isTypeParameter())
            return { kind: 0, type: { kind: 5, name: type.symbol.name } };
        else {
            const sig = this.extractor.checker.getSignaturesOfType(type, typescript_1.default.SignatureKind.Call)[0];
            if (sig)
                return {
                    kind: 1,
                    parameters: sig.parameters.map(p => ({
                        type: this.resolveTypeType(this.extractor.checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration)),
                        name: p.name,
                        rest: Boolean(p.declarations[0].dotDotDotToken)
                    })),
                    typeParameters: sig.typeParameters?.map(p => ({
                        name: p.symbol.name,
                        default: p.getDefault() ? this.resolveTypeType(p.getDefault()) : undefined,
                        constraint: p.getConstraint() ? this.resolveTypeType(p.getConstraint()) : undefined,
                    })),
                    returnType: this.resolveTypeType(sig.getReturnType())
                };
        }
        const typeStr = this.extractor.checker.typeToString(type);
        if (typeStr === "true")
            return { kind: 16 };
        else if (typeStr === "false")
            return { kind: 17 };
        else if (typeStr === "object")
            return { kind: 33 };
        const sym = type.getSymbol();
        if (sym)
            return this.resolveSymbol(sym, this.extractor.checker.getTypeArguments(type).map(t => this.resolveTypeType(t)));
        return { kind: 9, name: typeStr };
    }
    resolveAliasedSymbol(symbol) {
        while ((0, utils_1.hasBit)(symbol.flags, typescript_1.default.SymbolFlags.Alias)) {
            const newSym = this.extractor.checker.getAliasedSymbol(symbol);
            if (newSym.name === "unknown")
                return symbol;
            symbol = newSym;
        }
        return symbol;
    }
    getJSDocCommentOfParam(node) {
        const tag = typescript_1.default.getJSDocParameterTags(node)[0];
        if (!tag)
            return;
        return typescript_1.default.getTextOfJSDocComment(tag.comment);
    }
    resolveSymbolFileName(sym) {
        return sym.name.slice(1, -1) + ".ts";
    }
    getJSDocData(node) {
        const jsDoc = node.jsDoc;
        if (!jsDoc || !jsDoc.length)
            return undefined;
        return this.jsDocToJsDocData(jsDoc);
    }
    getJSDocDataRaw(node) {
        if (node.jsDoc)
            return this.getJSDocData(node);
        if (node.pos === -1)
            return;
        const fullText = node.getSourceFile().text;
        const ranges = typescript_1.default.getLeadingCommentRanges(fullText, node.getFullStart());
        if (!ranges)
            return;
        const jsDoc = [];
        for (const range of ranges) {
            const text = fullText.slice(range.pos, range.end);
            const parsed = typescript_1.default.parseIsolatedJSDocComment(text);
            if (parsed && parsed.jsDoc)
                jsDoc.push(parsed);
        }
        return this.jsDocToJsDocData(jsDoc);
    }
    jsDocToJsDocData(jsDoc) {
        const res = [];
        for (const currentDoc of jsDoc) {
            const tags = [];
            if (currentDoc.tags) {
                for (const tag of currentDoc.tags) {
                    tags.push({
                        name: tag.tagName.text,
                        comment: typescript_1.default.getTextOfJSDocComment(tag.comment),
                        arg: tag.name?.text,
                        type: tag.typeExpression && this.resolveType(tag.typeExpression.type)
                    });
                }
            }
            res.push({ comment: typescript_1.default.getTextOfJSDocComment(currentDoc.comment), tags });
        }
        return res;
    }
    getLOC(currentModule, node, includeFilename) {
        const sourceFile = node.getSourceFile();
        const pos = sourceFile.getLineAndCharacterOfPosition(node.getStart());
        if (!currentModule.repository)
            return { pos, filename: includeFilename ? (0, utils_1.getLastItemFromPath)(sourceFile.fileName) : undefined };
        if (currentModule.isNamespace)
            return { pos, sourceFile: `${currentModule.repository.slice(0, currentModule.repository.indexOf("#"))}#L${pos.line + 1}`, filename: includeFilename ? (0, utils_1.getLastItemFromPath)(sourceFile.fileName) : undefined };
        const filename = (0, utils_1.getLastItemFromPath)(sourceFile.fileName);
        return {
            pos,
            sourceFile: currentModule.repository && `${currentModule.repository}/${filename}#L${pos.line + 1}`,
            filename: includeFilename ? filename : undefined
        };
    }
    moduleToJSON(module = this.module) {
        const clone = { ...module };
        clone.modules = [];
        for (const [, mod] of module.modules) {
            clone.modules.push(this.moduleToJSON(mod));
        }
        return clone;
    }
    toJSON() {
        return {
            readme: this.readme,
            repository: this.repository,
            homepage: this.homepage,
            module: this.moduleToJSON()
        };
    }
    isInternalNode(node) {
        return this.extractor.settings.stripInternal && node.jsDoc?.some(t => t.tags && t.tags.some(tag => tag.tagName.text === "internal"));
    }
}
exports.Project = Project;
